import torch
from torch import Tensor
import torch.nn as nn
import torch.nn.functional as F

try:
    from torch.hub import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url
from typing import Type, Any, Callable, Union, List, Optional

__all__ = ['ResNet', 'resnet18', 'resnet34', 'resnet50', 'resnet101',
           'resnet152', 'resnext50_32x4d', 'resnext101_32x8d',
           'wide_resnet50_2', 'wide_resnet101_2']

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-f37072fd.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-b627a593.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-0676ba61.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-63fe2227.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-394f9c45.pth',
    'resnext50_32x4d': 'https://download.pytorch.org/models/resnext50_32x4d-7cdf4587.pth',
    'resnext101_32x8d': 'https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth',
    'wide_resnet50_2': 'https://download.pytorch.org/models/wide_resnet50_2-95faca4d.pth',
    'wide_resnet101_2': 'https://download.pytorch.org/models/wide_resnet101_2-32ee1156.pth',
}


def conv3x3(in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1) -> nn.Conv2d:
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(
            self,
            inplanes: int,
            planes: int,
            stride: int = 1,
            downsample: Optional[nn.Module] = None,
            groups: int = 1,
            base_width: int = 64,
            dilation: int = 1,
            norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion: int = 4

    def __init__(
            self,
            inplanes: int,
            planes: int,
            stride: int = 1,
            downsample: Optional[nn.Module] = None,
            groups: int = 1,
            base_width: int = 64,
            dilation: int = 1,
            norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out

class NormalizeLayer(nn.Module):
    """
    In order to certify radii in original coordinates rather than standardized coordinates, we
    add the Gaussian noise _before_ standardizing, which is why we have standardization be the first
    layer of the classifier rather than as a part of preprocessing as is typical.
    """

    def __init__(self):
        super(NormalizeLayer, self).__init__()

    def forward(self, inputs):
        return (inputs - 0.5) / 0.5

class ResNet(nn.Module):

    def __init__(
            self,
            block: Type[Union[BasicBlock, Bottleneck]],
            layers: List[int],
            pre=False,

            num_classes: int = 1000,
            zero_init_residual: bool = False,
            groups: int = 1,
            width_per_group: int = 64,
            replace_stride_with_dilation: Optional[List[bool]] = None,
            norm_layer: Optional[Callable[..., nn.Module]] = None,
            pretrain_weghts=False,
            is_wide=False
    ) -> None:
        super(ResNet, self).__init__()
        self.pre = pre
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups  # 1
        self.base_width = width_per_group  # 64
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2,
                                       dilate=replace_stride_with_dilation[0])
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2,
                                       dilate=replace_stride_with_dilation[1])
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2,
                                       dilate=replace_stride_with_dilation[2])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.pretrain_weghts = pretrain_weghts
        end_size = 2048 if is_wide else 512

        if not pre:
            self.fc = nn.Linear(end_size * block.expansion, num_classes)
        else:
            self.linear = nn.Linear(end_size * 1, 2)
            self.simclr_layer = nn.Sequential(
                    nn.Linear(end_size, end_size),
                    nn.ReLU(),
                    nn.Linear(end_size, 128)
                )
            self.shift_cls_layer = nn.Linear(end_size * 1, 2)
            self.joint_distribution_layer = nn.Linear(end_size * 1, 8)


        mu = torch.tensor([0.485, 0.456, 0.406]).view(3, 1, 1).cuda()
        std = torch.tensor([0.229, 0.224, 0.225]).view(3, 1, 1).cuda()
        self.norm = lambda x: (x - mu) / std
        self.normalize = None
        if pretrain_weghts:
            self.normalize = lambda x: (x - mu) / std
        else:
            self.normalize = NormalizeLayer()
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(self, block: Type[Union[BasicBlock, Bottleneck]], planes: int, blocks: int,
                    stride: int = 1, dilate: bool = False) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def _forward_impl(self, x: Tensor, eval_pre=False) -> Tensor:
        # See note [TorchScript super()]
        # x = self.norm(x)
        x = self.normalize(x)
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        feature_a = self.layer1(x)
        feature_b = self.layer2(feature_a)
        feature_c = self.layer3(feature_b)

        if eval_pre:
            feature_d = self.layer4(feature_c)
            feature_d = self.avgpool(feature_d)
            feature_d = torch.flatten(feature_d, 1)
            if self.pretrain_weghts:
                feature_d = F.normalize(feature_d, dim=1)
            return self.shift_cls_layer(feature_d)
        
        return [feature_a, feature_b, feature_c]

    def forward(self, x: Tensor, head_end=False, eval_pre=False) -> Tensor:
        if not head_end:
            return self._forward_impl(x, eval_pre)
        x = self.normalize(x)
        # x = self.norm(x)
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        feature_a = self.layer1(x)
        feature_b = self.layer2(feature_a)
        feature_c = self.layer3(feature_b)
        # print('feature_c:', feature_c.shape)
        # print('self.layer4:', self.layer4)
        feature_d = self.layer4(feature_c)
        if eval_pre:
            feature_d = self.avgpool(feature_d)
            feature_d = torch.flatten(feature_d, 1)
            if self.pretrain_weghts:
                feature_d = F.normalize(feature_d, dim=1)
            return self.shift_cls_layer(feature_d)
        # print('en out:', feature_a.shape, feature_b.shape, feature_c.shape)
        return [feature_a, feature_b, feature_c, feature_d]


def _resnet(
        arch: str,
        block: Type[Union[BasicBlock, Bottleneck]],
        layers: List[int],
        pretrained: bool,
        progress: bool,
        pre_path=None,
        head_end=False,
        is_pre_model=False,
        pretrain_weghts=False,
        **kwargs: Any
) -> ResNet:
    pre = True if pre_path is not None else False
    is_wide = True if arch == 'wide_resnet50_2' else False
    model = ResNet(block, layers, is_pre_model, pretrain_weghts=pretrain_weghts, is_wide=is_wide, **kwargs)
    if (not head_end) and (not is_pre_model):
        model.layer4 = None
        model.fc = None
    if pretrained:
        if pre:
            dic = torch.load(pre_path)
            # print('loaded keys:', dic.keys())
            # print('model:', model)
            key_list = list(dic.keys())
            # print(model_dic.keys())
            model.load_state_dict(dic)
            print("Pre is loaded!")
        else:
            state_dict = load_state_dict_from_url(model_urls[arch],
                                                  progress=progress)
            # for k,v in list(state_dict.items()):
            #    if 'layer4' in k or 'fc' in k:
            #        state_dict.pop(k)
            model.load_state_dict(state_dict)
    return model


class BN_layer(nn.Module):
    def __init__(self,
                 block: Type[Union[BasicBlock, Bottleneck]],
                 layers: int,
                 groups: int = 1,
                 width_per_group: int = 64,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 ):
        super(BN_layer, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer
        self.groups = groups
        self.base_width = width_per_group
        self.inplanes = 256 * block.expansion
        self.dilation = 1
        self.bn_layer = self._make_layer(block, 512, layers, stride=2)

        self.conv1 = conv3x3(64 * block.expansion, 128 * block.expansion, 2)
        self.bn1 = norm_layer(128 * block.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(128 * block.expansion, 256 * block.expansion, 2)
        self.bn2 = norm_layer(256 * block.expansion)
        self.conv3 = conv3x3(128 * block.expansion, 256 * block.expansion, 2)
        self.bn3 = norm_layer(256 * block.expansion)

        self.conv4 = conv1x1(1024 * block.expansion, 512 * block.expansion, 1)
        self.bn4 = norm_layer(512 * block.expansion)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block: Type[Union[BasicBlock, Bottleneck]], planes: int, blocks: int,
                    stride: int = 1, dilate: bool = False) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes * 3, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes * 3, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def forward(self, x):
        # See note [TorchScript super()]
        # x = self.cbam(x)
        # print('bn input shape:', [y.shape for y in x])
        l1 = self.relu(self.bn2(self.conv2(self.relu(self.bn1(self.conv1(x[0]))))))
        l2 = self.relu(self.bn3(self.conv3(x[1])))
        feature = torch.cat([l1, l2, x[2]], 1)
        output = self.bn_layer(feature)
        # print('bn output shape:', output.shape)
        # x = self.avgpool(feature_d)
        # x = torch.flatten(x, 1)
        # x = self.fc(x)

        return output.contiguous()


def resnet18(pretrained: bool = False, progress: bool = True, pre_path=None, head_end=False, is_pre_model=False, pretrain_weghts=False, **kwargs: Any):
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    encoder = _resnet('resnet18', BasicBlock, [2, 2, 2, 2], pretrained, progress, pre_path=pre_path, head_end=head_end, is_pre_model=is_pre_model, pretrain_weghts=pretrain_weghts,
                      **kwargs)
    if 'norm_layer' in kwargs:
        kwargs.pop('norm_layer')
    bn = BN_layer(BasicBlock, 2, **kwargs)
    return encoder, bn


def resnet34(pretrained: bool = False, progress: bool = True, pretrain_weghts=False, **kwargs: Any):
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    encoder = _resnet('resnet34', BasicBlock, [3, 4, 6, 3], pretrained, progress, pretrain_weghts=pretrain_weghts,
                      **kwargs)
    if 'norm_layer' in kwargs:
        kwargs.pop('norm_layer')
    bn = BN_layer(BasicBlock, 3, **kwargs)
    return encoder, bn


def resnet50(pretrained: bool = False, pretrain_weghts=False, progress: bool = True, **kwargs: Any):
    r"""ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    encoder = _resnet('resnet50', Bottleneck, [3, 4, 6, 3],
                      pretrained, progress, pretrain_weghts=pretrain_weghts, **kwargs)
    if 'norm_layer' in kwargs:
        kwargs.pop('norm_layer')
    bn = BN_layer(Bottleneck, 3, **kwargs)
    return encoder, bn


def resnet101(pretrained: bool = False, pretrain_weghts=False, progress: bool = True, **kwargs: Any):
    r"""ResNet-101 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """

    encoder = _resnet('resnet101', Bottleneck, [3, 4, 23, 3],
                      pretrained, progress, pretrain_weghts=pretrain_weghts, **kwargs)
    if 'norm_layer' in kwargs:
        kwargs.pop('norm_layer')
    bn = BN_layer(Bottleneck, 3, **kwargs)
    return encoder, bn


def resnet152(pretrained: bool = False, pretrain_weghts=False, progress: bool = True, **kwargs: Any):
    r"""ResNet-152 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    encoder = _resnet('resnet152', Bottleneck, [3, 8, 36, 3],
                      pretrained, progress, pretrain_weghts=pretrain_weghts, **kwargs)
    if 'norm_layer' in kwargs:
        kwargs.pop('norm_layer')
    bn = BN_layer(Bottleneck, 3, **kwargs)
    return encoder, bn


def wide_resnet50_2(pretrained: bool = False, progress: bool = True, pre_path=None, head_end=False, is_pre_model=False, pretrain_weghts=False, **kwargs: Any):
    r"""Wide ResNet-50-2 model from
    `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_.
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs['width_per_group'] = 64 * 2
    encoder = _resnet('wide_resnet50_2', Bottleneck, [3, 4, 6, 3],
                      pretrained, progress, pre_path=pre_path, head_end=head_end, is_pre_model=is_pre_model, pretrain_weghts=pretrain_weghts, **kwargs)
    if 'norm_layer' in kwargs:
        kwargs.pop('norm_layer')
    bn = BN_layer(Bottleneck, 3, **kwargs)
    return encoder, bn


def wide_resnet101_2(pretrained: bool = False, progress: bool = True, pretrain_weghts=False, **kwargs: Any):
    r"""Wide ResNet-101-2 model from
    `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_.
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs['width_per_group'] = 64 * 2
    encoder = _resnet('wide_resnet101_2', Bottleneck, [3, 4, 23, 3],
                      pretrained, progress, pretrain_weghts=pretrain_weghts, **kwargs)
    if 'norm_layer' in kwargs:
        kwargs.pop('norm_layer')
    bn = BN_layer(Bottleneck, 3, **kwargs)
    return encoder, bn


def resnext50_32x4d(pretrained: bool = False, progress: bool = True, pretrain_weghts=False, **kwargs: Any):
    r"""ResNeXt-50 32x4d model from
    `"Aggregated Residual Transformation for Deep Neural Networks" <https://arxiv.org/pdf/1611.05431.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs['groups'] = 32
    kwargs['width_per_group'] = 4

    encoder = _resnet('resnext50_32x4d', Bottleneck, [3, 4, 6, 3],
                      pretrained, progress, pretrain_weghts=pretrain_weghts, **kwargs)
    if 'norm_layer' in kwargs:
        kwargs.pop('norm_layer')
    bn = BN_layer(Bottleneck, 3, **kwargs)
    return encoder, bn
